#!/usr/bin/env python

import collections
import hashlib
import json
import os
import psutil
import smbus
import sqlite3
import subprocess
import sys
import threading
import time
import yaml

from pysnmp import debug
from pysnmp.carrier.asyncore.dgram import udp
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.proto.api import v2c
from pysnmp.smi import builder, instrum, exval, compiler
from usbid import USB

MibObject = collections.namedtuple('MibObject', ['mibName', 'objectType', 'valueGetFunc', 'valueSetFunc'])
def createVariable(SuperClass, getValue, setValue, *args):
    class Var(SuperClass):
        def readGet(self, name, *args):
            return name, self.syntax.clone(getValue())
        def writeTest(self, name, *args ):
            pass
        def writeCommit(self, name, val, *args ):
            setValue(val)
    return Var(*args)
class NIOEUSBInfo():
    def __init__(self):
        self._lock = threading.RLock()
        self.niousbConfigPath = '/etc/niousb/niousb.yml'
        self.configLastModified = 0
        self.prevConfigLastModified = 0
    def getUSBInfo(self):
        self.dictUSB = {}
        def usbPort(port, list):
            for p in port.values():
                list.append(p)
                usbPort(p, list)
        usb = USB()
        usbPortsList = []
        for u in usb.keys():
            bus = usb[u]
            for b in bus.keys():
                port = bus[b]
                usbPort(port, usbPortsList)
        for i in usbPortsList:
            devPath = str(i.fs_path[i.fs_path.rfind(os.path.sep) + 1:])
            idProduct = str(getattr(i, 'idProduct', ''))
            idVendor = str(getattr(i, 'idVendor', ''))
            product = str(getattr(i, 'product', ''))
            manufacturer = str(getattr(i, 'manufacturer', ''))
            self.dictUSB.update(dict({devPath: [idProduct, idVendor, product, manufacturer]}))
    def getValues(self):
        with self._lock:
            self.getUSBInfo()
            self.configLastModified = int(os.stat(self.niousbConfigPath).st_mtime)
            if(self.configLastModified != self.prevConfigLastModified):
                self.niousbConfig = yaml.load(open(self.niousbConfigPath).read(), Loader=yaml.SafeLoader)
                self.prevConfigLastModified = self.configLastModified
            result = []
            clusterSelect = -1
            try:
                clusterSelect = int(self.niousbConfig['HAConfiguration']['ClusterSelect'])
            except:
                pass
            try:
                index = 0
                for pcb in self.niousbConfig['PCBConfiguration']:
                    if pcb['i2caddress'] != clusterSelect: 
                        if 'order' in pcb:
                            hub = int(pcb['order'])
                        else:
                            hub = -1
                        for port in pcb['ports']:
                            index += 1
                            if 'devPath' in port:
                                try:
                                    idProduct = self.dictUSB[port['devPath']][0]
                                    idVendor = self.dictUSB[port['devPath']][1]
                                    product = self.dictUSB[port['devPath']][2]
                                    manufacturer = self.dictUSB[port['devPath']][3]
                                except:
                                    idProduct = ''
                                    idVendor = ''
                                    product = ''
                                    manufacturer = ''
                            if 'num' in port:
                                num = int(port['num'])
                            else:
                                num = -1
                            if 'powerState' in port:
                                powerState = int(port['powerState'])
                            else:
                                powerState = -1
                            if 'nickname' in port:
                                nickname = port['nickname'].encode('utf-8')
                            else:
                                nickname = ''
                            if 'description' in port:
                                description = port['description'].encode('utf-8')
                            else:
                                description = ''
                            resultPort = [index, hub, num, powerState, nickname, description, idProduct, idVendor, product, manufacturer]
                            result.append(resultPort)
            except:
                pass
            return result
class CustomMib(object):
    def __init__(self):
        self._lock = threading.RLock()
        self._systemVoltage = -1
        self._psu1Voltage = -1
        self._psu2Voltage = -1
        self._temp = -1
        self._dbPath = '/tmp/.hw.db'
    def getIsMasterNode(self):
        with self._lock:
            return int(os.popen('/sbin/ifconfig | /bin/grep -c "eth0.*:0\\b"').read(), 10)
    def getSystemVoltage(self):
        with self._lock:
            try:
                conn = sqlite3.connect(self._dbPath)
                cursor = conn.cursor()
                sql = "SELECT systemVoltage FROM adc1115 ORDER BY id DESC LIMIT 1"
                cursor.execute(sql)
                values = cursor.fetchone()
                if values:
                    self._systemVoltage = values[0] * 1000
                else:
                    self._systemVoltage = -1
            except:
                self._systemVoltage = -1
            return int(self._systemVoltage)
    def getPsu1Voltage(self):
        with self._lock:
            try:
                conn = sqlite3.connect(self._dbPath)
                cursor = conn.cursor()
                sql = "SELECT psu1Voltage FROM adc1115 ORDER BY id DESC LIMIT 1"
                cursor.execute(sql)
                values = cursor.fetchone()
                if values:
                    self._psu1Voltage = values[0] * 1000
                else:
                    self._psu1Voltage = -1
            except:
                self._psu1Voltage = -1
            return int(self._psu1Voltage)
    def getPsu2Voltage(self):
        with self._lock:
            try:
                conn = sqlite3.connect(self._dbPath)
                cursor = conn.cursor()
                sql = "SELECT psu2Voltage FROM adc1115 ORDER BY id DESC LIMIT 1"
                cursor.execute(sql)
                values = cursor.fetchone()
                if values:
                    self._psu2Voltage = values[0] * 1000
                else:
                    self._psu2Voltage = -1
            except:
                self._psu2Voltage = -1
            return int(self._psu2Voltage)
    def getTemp(self):
        with self._lock:
            try:
                conn = sqlite3.connect(self._dbPath)
                cursor = conn.cursor()
                sql = "SELECT temp FROM adc1115 ORDER BY id DESC LIMIT 1"
                cursor.execute(sql)
                values = cursor.fetchone()
                if values:
                    self._temp = values[0] * 100
                else:
                    self._temp = -1
            except:
                self._temp = -1
            return int(round(self._temp))
    def getCpuTotalUsage(self):
        with self._lock:
            return psutil.cpu_percent()
    def getMemTotalReal(self):
        with self._lock:
            return psutil.virtual_memory().total
    def getMemTotalFree(self):
        with self._lock:
            return psutil.virtual_memory().available
    def getUptime(self):
        with self._lock:
            return int(time.time() - psutil.boot_time())
class SNMPAgent(object):
    def __init__(self, mibObjects, community='public'):
        self._snmpEngine = engine.SnmpEngine()
        config.addSocketTransport(self._snmpEngine, udp.domainName, udp.UdpTransport().openServerMode(('0.0.0.0', 161)))
        config.addV1System(self._snmpEngine, "read-area", community)
        config.addVacmUser(self._snmpEngine, 1, "read-area", "noAuthNoPriv", readSubTree=(1, 3, 6, 1, 4, 1, 53034))
        config.addVacmUser(self._snmpEngine, 2, "read-area", "noAuthNoPriv", readSubTree=(1, 3, 6, 1, 4, 1, 53034))
        self._snmpContext = context.SnmpContext(self._snmpEngine)
        self.mibBuilder = self._snmpContext.getMibInstrum().getMibBuilder()
        mibSources = self.mibBuilder.getMibSources() + (builder.DirMibSource('.'),) + (builder.DirMibSource('./pysnmp_mibs'),)
        self.mibBuilder.setMibSources(*mibSources)
        MibScalarInstance, = self.mibBuilder.importSymbols('SNMPv2-SMI', 'MibScalarInstance')
        for mibObject in mibObjects:
            nextVar, = self.mibBuilder.importSymbols(mibObject.mibName, mibObject.objectType)
            instance = createVariable(MibScalarInstance, mibObject.valueGetFunc, mibObject.valueSetFunc, nextVar.name, (0,), nextVar.syntax)
            instanceDict = {str(nextVar.name) + "Instance":instance}
            self.mibBuilder.exportSymbols(mibObject.mibName, **instanceDict)
        cmdrsp.GetCommandResponder(self._snmpEngine, self._snmpContext)
        cmdrsp.SetCommandResponder(self._snmpEngine, self._snmpContext)
        cmdrsp.NextCommandResponder(self._snmpEngine, self._snmpContext)
        cmdrsp.BulkCommandResponder(self._snmpEngine, self._snmpContext)
    def serve_forever(self):
        self._snmpEngine.transportDispatcher.jobStarted(1)
        try:
            self._snmpEngine.transportDispatcher.runDispatcher()
        except:
            self._snmpEngine.transportDispatcher.closeDispatcher()
            raise
class Worker(threading.Thread):
    def __init__(self, agent, updateInterval=300):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.updateInterval = updateInterval
        self._agent = agent
        self.portsEntryTable, = self._agent.mibBuilder.importSymbols('NIO-EUSB-MIB', 'portsEntry')
        self.mibInstrum = self._agent._snmpContext.getMibInstrum()
        self.usbTable = NIOEUSBInfo()
    def run(self):
        while True:
            result = []
            value = self.usbTable.getValues()
            for a in range(len(value)):
                instanceId = self.portsEntryTable.getInstIdFromIndices(a + 1)
                for v in range(len(value[a])):
                    result.append((self.portsEntryTable.name + (v + 1,) + instanceId, value[a][v]))
            self.mibInstrum.writeVars(tuple(result))
            time.sleep(float(self.updateInterval))
if __name__ == '__main__':
    try:
        niousbConfig = yaml.load(open('/etc/niousb/niousb.yml').read(), Loader=yaml.SafeLoader)
        modelId = niousbConfig['DeviceConfiguration']['ModelId']
    except:
        modelId = ''
    try:
        snmpdConfig = yaml.load(open('/etc/niousb/snmpd.yml').read(), Loader=yaml.SafeLoader)
        snmpdCommunity = snmpdConfig['SNMPdConfiguration']['readCommunity']
        tableUpdateInterval = snmpdConfig['SNMPdConfiguration']['tableUpdateInterval']
    except:
        snmpdCommunity = 'public'
        tableUpdateInterval = 60
    customMib = CustomMib()
    if (modelId == '16epcl.b' or modelId == '16epcl.c' or modelId == '21ipcl.c' or modelId == '21ipcl.d'):
        objects = [MibObject('NIO-EUSB-MIB', 'isMasterNode', customMib.getIsMasterNode, ''),
                MibObject('NIO-EUSB-MIB', 'cpuTotalUsage', customMib.getCpuTotalUsage, ''),
                MibObject('NIO-EUSB-MIB', 'memTotalReal', customMib.getMemTotalReal, ''),
                MibObject('NIO-EUSB-MIB', 'memTotalFree', customMib.getMemTotalFree, ''),
                MibObject('NIO-EUSB-MIB', 'uptime', customMib.getUptime, ''),
                MibObject('NIO-EUSB-MIB', 'systemVoltage', customMib.getSystemVoltage, ''),
                MibObject('NIO-EUSB-MIB', 'psu1Voltage', customMib.getPsu1Voltage, ''),
                MibObject('NIO-EUSB-MIB', 'psu2Voltage', customMib.getPsu2Voltage, ''),
                MibObject('NIO-EUSB-MIB', 'temp', customMib.getTemp, ''),]
    elif (modelId == '21ip.b'):
        objects = [MibObject('NIO-EUSB-MIB', 'isMasterNode', customMib.getIsMasterNode, ''),
                MibObject('NIO-EUSB-MIB', 'cpuTotalUsage', customMib.getCpuTotalUsage, ''),
                MibObject('NIO-EUSB-MIB', 'memTotalReal', customMib.getMemTotalReal, ''),
                MibObject('NIO-EUSB-MIB', 'memTotalFree', customMib.getMemTotalFree, ''),
                MibObject('NIO-EUSB-MIB', 'uptime', customMib.getUptime, ''),
                MibObject('NIO-EUSB-MIB', 'systemVoltage', customMib.getSystemVoltage, ''),
                MibObject('NIO-EUSB-MIB', 'psu1Voltage', customMib.getPsu1Voltage, ''),
                MibObject('NIO-EUSB-MIB', 'temp', customMib.getTemp, ''),]
    else:
        objects = [MibObject('NIO-EUSB-MIB', 'isMasterNode', customMib.getIsMasterNode, ''),
                MibObject('NIO-EUSB-MIB', 'cpuTotalUsage', customMib.getCpuTotalUsage, ''),
                MibObject('NIO-EUSB-MIB', 'memTotalReal', customMib.getMemTotalReal, ''),
                MibObject('NIO-EUSB-MIB', 'memTotalFree', customMib.getMemTotalFree, ''),
                MibObject('NIO-EUSB-MIB', 'uptime', customMib.getUptime, ''),]
    agent = SNMPAgent(objects, snmpdCommunity)
    Worker(agent, tableUpdateInterval).start()
    try:
        agent.serve_forever()
    except:
        pass
